//https://github.com/urho3d/Urho3D/blob/master/bin/CoreData/Shaders/GLSL/BRDF.glsl
//https://google.github.io/filament/Filament.md.html#materialsystem/diffusebrdf
//https://schuttejoe.github.io/post/disneybsdf/

float schlick_fresnel(float VdotH, float ior, float f0){
    return (ior - f0) * pow5(1.0 - VdotH) + f0;
}

float schlick_gaussian_fresnel(float VdotH, float ior, float f0){
    return (ior - f0) * pow(2.0, (-5.55473 * VdotH - 6.98316) * VdotH) + f0;
}

float schlick_weight(float u){
    float m = saturate(1.0 - u);
    float m2 = m * m;

    return m * m2 * m2;
}

vec3 exact_fresnel(const vec3 n, const vec3 k, float c) {
    const vec3 k2= k * k;
	const vec3 n2k2 = n * n + k2;

    vec3 c2n = (c * 2.0) * n;
    vec3 c2 = vec3(c * c);

    vec3 rs_num = n2k2 - c2n + c2;
    vec3 rs_den = n2k2 + c2n + c2;

    vec3 rs = rs_num / rs_den;

    vec3 rp_num = n2k2 * c2 - c2n + 1;
    vec3 rp_den = n2k2 * c2 + c2n + 1;

    vec3 rp = rp_num / rp_den;

    return saturate(0.5 * (rs + rp));
}

vec3 f0_to_ior(vec3 f0){
    return (1.0 + f0) / (1.0 - f0);
}

vec3 fresnel(float theta, float ior, float f0, vec3 albedo, float metal){
    vec3 n, k;

    #if SPEC_FORMAT == 0
        if(f0 > 229.5 / 255.0){
            if(metal == 2){             //Fer
                n = vec3(2.9114, 2.9497, 2.5845);
                k = vec3(0.18299, 0.42108, 1.3734);
            } else if(metal == 3){      //Or
                n = vec3(0.18299, 0.42108, 1.3734);
                k = vec3(3.4242, 2.3459, 1.7704);
            } else if(metal == 4){      //Aluminium
                n = vec3(1.3456, 0.96521, 0.61722);
                k = vec3(7.4746, 6.3995, 5.3031);
            } else if(metal == 5){      //Chrome
                n = vec3(3.1071, 3.1812, 2.3230);
                k = vec3(3.3314, 3.3291, 3.1350);
            } else if(metal == 6){      //Cuivre
                n = vec3(0.27105, 0.67693, 1.3164);
                k = vec3(3.6092, 2.6248, 2.2921);
            } else if(metal == 7){      //Plomb
                n = vec3(1.9100, 1.8300, 1.4400);
                k = vec3(3.5100, 3.4000, 3.1800);
            } else if(metal == 8){      //Platine
                n = vec3(2.3757, 2.0847, 1.8453);
                k = vec3(4.2655, 3.7153, 3.1365);
            } else if(metal == 9){      //Silver
                n = vec3(0.15943, 0.14512, 0.13547);
                k = vec3(3.9291, 3.1900, 2.3808);
            //} else {
                //n = f0_to_ior(albedo);
                //k = vec3(0.0);
            }

            return exact_fresnel(n, k, theta);
        }
    #endif

    return vec3(schlick_fresnel(theta, ior, f0));
}

float ggx_g2(float NdotL, float NdotV, float roughness){
    float g0 = (2.0 * NdotV) / (NdotV + sqrt(roughness + (1.0 - roughness) * (NdotV * NdotV)));
    float g1 = (2.0 * NdotL) / (NdotL + sqrt(roughness + (1.0 - roughness) * (NdotL * NdotL)));

    return g0 * g1;
}

float d_ggx(float NdotH, float roughness){
    float tmp = (NdotH * roughness - NdotH) * NdotH + 1.0;
    return roughness / (tmp * tmp) * rpi;
}

float burley(float NdotL, float NdotV, float roughness, float ior){
    float f0 = 0.5 + 2.0 * roughness * NdotL * NdotL;
    float fl = schlick_weight(NdotL);
    float fv = schlick_weight(NdotV);

    return f0 * (fl + fv + fl * fv * (f0 - ior));
}

float disney_diffuse(float NdotL, float NdotV, float LdotH, float roughness, float ior){
    float fl = schlick_weight(NdotL);
    float fv = schlick_weight(NdotV);

    float fss90 = LdotH * LdotH * roughness;
    float fss = mix(1.0, fss90, fl) * mix(1.0, fss90, fv);
    float ss = 1.25 * (fss * (1.0 / (NdotL + NdotV) - 0.5) + 0.5);

    float burley = burley(NdotL, NdotV, roughness, ior);

    return rpi * (burley + (1.0 - 0.5 * fl) * (1.0 - 0.5 * fv));
}

#if defined DEFERRED || defined COMPOSITE
    float NdotH_squared(float radiusTan, float NoL, float NoV, float VoL){
        float radiusCos = inversesqrt(1.0 + radiusTan * radiusTan);
        
        float RoL = 2.0 * NoL * NoV - VoL;
        if (RoL >= radiusCos)
            return 1.0;

        float rOverLengthT = radiusCos * radiusTan * inversesqrt(1.0 - RoL * RoL);
        float NoTr = rOverLengthT * (NoV - RoL * NoL);
        float VoTr = rOverLengthT * (2.0 * NoV * NoV - 1.0 - RoL * VoL);

        float triple = sqrt(clamp(1.0 - NoL * NoL - NoV * NoV - VoL * VoL + 2.0 * NoL * NoV * VoL,0.,1.));
        
        float NoBr = rOverLengthT * triple, VoBr = rOverLengthT * (2.0 * triple * NoV);
        float NoLVTr = NoL * radiusCos + NoV + NoTr, VoLVTr = VoL * radiusCos + 1.0 + VoTr;
        float p = NoBr * VoLVTr, q = NoLVTr * VoLVTr, s = VoBr * NoLVTr;    
        float xNum = q * (-0.5 * p + 0.25 * VoBr * NoLVTr);
        float xDenom = p * p + s * ((s - 2.0 * p)) + NoLVTr * ((NoL * radiusCos + NoV) * VoLVTr * VoLVTr + 
                    q * (-0.5 * (VoLVTr + VoL * radiusCos) - 0.5));
        float twoX1 = 2.0 * xNum / (xDenom * xDenom + xNum * xNum);
        float sinTheta = twoX1 * xDenom;
        float cosTheta = 1.0 - twoX1 * xNum;
        NoTr = cosTheta * NoTr + sinTheta * NoBr;
        VoTr = cosTheta * VoTr + sinTheta * VoBr;
        
        float newNoL = NoL * radiusCos + NoTr;
        float newVoL = VoL * radiusCos + VoTr;
        float NoH = NoV + newNoL;
        float HoH = 2.0 * newVoL + 2.0;
        return max(0.0, NoH * NoH / HoH);
    }

    float ggx_hightligh(materials m, vec3 normal, vec3 vector, vec3 direction, float roughness, float f0){
        atmosphere_constant ac = atmosphere();

        #if defined DEFERRED6
            normal = m.grass ? vec3(0.0, 1.0, 0.0) : normal;
        #endif

        vec3 H = normalize(direction - vector);

        float radius = sunAngle < 0.5 ? ac.sun_angular_radius : ac.moon_angular_radius;

        float NdotL = dot(normal, direction);
        float NdotV = dot(normal, -vector);
        float LdotV = dot(direction, -vector);
        float LdotH = dot(direction, H);
        float NdotH = NdotH_squared(radius, NdotL, NdotV, LdotV);

        float d = d_ggx(NdotH, roughness);
        float g = ggx_g2(NdotL, NdotV, roughness);
        float f = schlick_gaussian_fresnel(LdotH, m.ior, f0);

        return saturate(NdotL) * (f * d * g / (4.0 * NdotL * NdotV));
    }
#endif