#include "/program/shaders1/common/specular/raytrace.glsl"

mat3 tbn(vec3 n){
    vec3 t = normalize(cross(gbufferModelView[1].xyz, n));
    vec3 b = cross(n, t);

    return mat3(t, b, n);
}

vec3 sample_specular(float i, float alpha2, const int steps) {
    float x = (alpha2 * i) / (1 - i);
    float y = i * float(steps) * pow2(16) * golden_angle;

    float c = inversesqrt(x + 1);
    float s = sqrt(x) * c;

    return vec3(cos(y) * s, sin(y) * s, c);
}

vec2 hammersley(uint i, uint numSamples) {
    uint bits = i;
    bits = (bits << 16) | (bits >> 16);
    bits = ((bits & 0x55555555) << 1) | ((bits & 0xAAAAAAAA) >> 1);
    bits = ((bits & 0x33333333) << 2) | ((bits & 0xCCCCCCCC) >> 2);
    bits = ((bits & 0x0F0F0F0F) << 4) | ((bits & 0xF0F0F0F0) >> 4);
    bits = ((bits & 0x00FF00FF) << 8) | ((bits & 0xFF00FF00) >> 8);
    return vec2(i / numSamples, bits / exp2(32));
}

vec3 GGXVNDF(vec3 vector, vec2 offset, float alpha){
    vec3 v = normalize(vec3(alpha * vector.x, alpha * vector.y, vector.z));

    vec3 t0 = v.z < 0.9999 ? normalize(cross(v, vec3(0.0, 0.0, 1.0))) : vec3(1.0, 0.0, 0.0);
    vec3 t1 = cross(t0, v);

    float a = 1.0 / (1.0 + v.z);
    float r = sqrt(offset.x);
    float p = offset.y < a ? offset.y / a * pi : pi + (offset.y - a) / (1.0 - a) * pi;

    float p0 = r * cos(p);
    float p1 = r * sin(p) * (offset.y < a ? 1.0 : v.z);

    vec3 n = p0 * t0 + p1 * t1 + sqrt(max0(1.0 - p0 * p0 - p1 * p1)) * v;
         n = normalize(vec3(alpha * n.x, alpha * n.y, max0(n.z)));
         
    return n;
}

vec3 atmosphere_reflection(vec3 position, vec3 moon_position){
    const float lod = ATMOSPHERE_LOD;

    float clouds_transmittance = texture2(colortex9, project_sphere(position) / lod).w;

    vec3 scattering = texture2(colortex9, project_sphere(position) / lod).xyz;
    vec3 transmittance = texture2(colortex7, project_sphere(position) / lod).xyz * pow4(clouds_transmittance);

    vec3 atmosphere = scattering;
         atmosphere += stars(position, moon_position) * luminance(transmittance);

    return atmosphere;
}

vec3 reflection(materials m, vec3 color, vec3 view_position, vec3 position, vec3 vector, vec3 normal, vec3 albedo, float depth, float lightmap, float roughness, float f0){
    const int steps0 = REFLECTION_QUALITY;
    const uint steps1 = 2;

    const float z_depth = 0.05;

    float alpha = pow2(roughness);
    float NdotV = max(1e-7, dot(normal, -vector));

    vec4 noise = blue_noise(gl_FragCoord.xy);
        
    vec3 reflection = vec3(0.0);
    vec3 hit_position = vec3(0.0);

    vec3 sun_position = mat3(gbufferModelViewInverse) * normalize(sunPosition);
    vec3 moon_position = mat3(gbufferModelViewInverse) * normalize(moonPosition);

    lightmap = saturate(lightmap);
    lightmap = 1.0 - pow(1.0 - lightmap, 0.75);
    lightmap = pow2(lightmap);

    if(depth >= 1) return color;
    if(roughness == 1) return color;

    for(uint i = 0; i < steps1; i++){
        vec2 xi = hammersley(i, steps1) + noise.xy;
        vec3 h = tbn(normal) * GGXVNDF(-vector * tbn(normal), xi, alpha);
        vec3 l = (2.0 * dot(h, -vector)) * h + vector;

        float LdotH = saturate(dot(l, h));
        float NdotL = saturate(dot(normal, l));
        #ifdef TERRAIN_REFLECTION
            bool intersect = screenspace_raytrace(view_position, position, mat3(gbufferModelView) * l, NdotV, noise.w, z_depth, steps0, hit_position);

            if(intersect){
                if(isEyeInWater == 0) reflection += decodeRGBE8(texture2(colortex0, hit_position.xy));
            } else {
                if(isEyeInWater == 0){
                    reflection += atmosphere_reflection(l, moon_position) * lightmap;
                }
            }
        #else
            if(isEyeInWater == 0){
                reflection += atmosphere_reflection(l, moon_position) * lightmap;
            }
        #endif

        reflection *= fresnel(LdotH, m.ior, f0, albedo, m.metal);
        reflection *= ggx_g2(NdotL, NdotV, alpha * alpha);
    }

    reflection /= float(steps1);
    
    if(f0 > 229.5 / 255 || m.stained_glass || m.slime){
        reflection *= (1.0 - f0) + f0 * albedo;
    }

    return color + max0(reflection);
}