#if defined COMPOSITE0
    float volumetric_fog_fbm(vec3 position, vec3 direction, const int octaves){
        float frequency = 3.75;
        float amplitude = 0.5;

        float noise = noise_3D(position);

        for(int i = 0; i < octaves; i++){
            noise += amplitude * noise_3D(position);
            position = position * frequency + direction;
            amplitude *= 0.5;
        }

        return noise;
    }

    float volumetric_fog_density(atmosphere_constant ac, volumetric_fog_settings vf, vec3 position){
        position += cameraPosition;

        float animation = frameTimeCounter * vf.speed * pow2(rpi);
        float altitude = (position.y - vf.altitude_min) / vf.thickness;

        float attenuation0 = saturate(remap(altitude, 0, 0.4, 0, 1) * remap(altitude, 0.6, 1, 1, 0));
        float attenuation1 = saturate(remap(altitude, 0, 0.2, 0, 1) * remap(altitude, 0.8, 1, 1, 0));

        vec3 direction0 = animation * normalize(vec3(-1.0, 0.0, 1.0));
        vec3 direction1 = animation * normalize(vec3(-1.0, 0.5, 1.0));

        vec3 position0 = position * vec3(1.0, 0.0, 1.0) * 0.025 + direction0;
        vec3 position1 = position * 0.2 + direction1;

        float noise0 = volumetric_fog_fbm(position0, direction0, vf.octaves0);
        float noise1 = volumetric_fog_fbm(position1, direction1, vf.octaves1) * 0.5;

        float noise = curve(noise0 * vf.coverage - noise1) * 0.5 + 0.5;
              noise *= exp2(-(position.y - vf.altitude_min) * 0.065);

        return saturate(noise * 2.0 - 1.0) * vf.density;
    }

    float volumetric_fog_transmittance(atmosphere_constant ac, volumetric_fog_settings vf, vec3 position, vec3 direction, const int steps){
        float ray_length = 22.0;
        float transmittance = 0.0;

        for(int i = 0; i < steps; i++, position += direction * ray_length){
            transmittance += volumetric_fog_density(ac, vf, position) * ray_length;
        }

        return transmittance;
    }

    vec4 volumetric_fog(vec3 position0, vec3 position1, vec3 direction, vec3 sun_position, vec3 moon_position, vec4 noise){
        atmosphere_constant ac = atmosphere();
        volumetric_fog_settings vf = s_volume_fog();

        vec3 direct, indirect;
        float transmittance = 1.0;

        float VdotL = dot(f_normalize(position1), direction);
        float VdotU = dot(f_normalize(position1), vec3(0.0, 1.0, 0.0));

        float phase0 = rayleigh_phase(VdotL);
        float phase1 = hg_phase(VdotL, ac.mie_g);

        #ifdef VOLUMETRIC_FOG_MORNING
            float sunrise = saturate(dot(sun_position, vec3(1.0, 0.0, 0.0)));
                  sunrise = pow4(sunrise * 0.7);
        #else
            float sunrise = 1.0;
        #endif

        position0 = clouds_curvature(position0);
        position1 = clouds_curvature(position1);

        vec3 ray_increment = (position1 - position0) / vf.steps0;
        vec3 ray_position = ray_increment * noise.w + position0;

        float ray_length = distance(position0, position1) / 16.0;

        vec3 shadow_position0 = shadow_position(position0);
        vec3 shadow_position1 = shadow_position(position1);

        vec3 shadow_increment = (shadow_position1 - shadow_position0) / vf.steps0;
        vec3 shadow_position = shadow_increment * noise.w + shadow_position0;
    
        for(int i = 0; i < vf.steps0; i++, ray_position += ray_increment, shadow_position += shadow_increment){
            if(transmittance <= vf.treshold) break;

            float density = volumetric_fog_density(ac, vf, ray_position) * ray_length * sunrise;
            if(density <= 0.0) continue;

            float step_transmittance = exp2(-density);
            float step_transmittance_fraction = 1.0 - step_transmittance;

            float direct_transmittance = volumetric_fog_transmittance(ac, vf, ray_position, direction, 4);
            float indirect_transmittance = volumetric_fog_transmittance(ac, vf, ray_position, vec3(0.0, 1.0, 0.0), 2) * 0.1;

            float powder = 1.0 - exp(-2.0 * density);

            float direct_powder = mix(powder, 1.0, VdotL * 0.5 + 0.5);
            float indirect_powder = mix(powder, 1.0, VdotU * 0.5 + 0.5);

            direct_powder = mix(direct_powder, 2.0, mie_phase(VdotL, ac.mie_g));

            vec3 shadow = volumetric_shadow(ray_position, shadow_position, false);

            for(int j = 0; j < vf.steps1; j++){
                float a = pow(0.6, j);
                float b = pow(0.4, j);
                float c = pow(0.8, j);

                float sample_transmittance = transmittance * (step_transmittance_fraction * a);

                float step_direct_transmittance = sample_transmittance * exp(-direct_transmittance * b);
                float step_indirect_transmittance = sample_transmittance * exp2(-indirect_transmittance * b);

                float direct_phase = clouds_phase(VdotL * c, direct_transmittance);
                float indirect_phase = 2.0 * clouds_phase(VdotU * c, indirect_transmittance);

                direct += step_direct_transmittance * direct_phase * direct_powder * shadow;
                indirect += step_indirect_transmittance * indirect_phase * indirect_powder;
            }

            transmittance *= step_transmittance;
        }
        
        transmittance = saturate((transmittance - vf.treshold) / (1.0 - vf.treshold));
    
        float ambient_occlusion = eyeBrightnessSmooth.y * (1.0 / 254.0);
              ambient_occlusion = pow2(ambient_occlusion);

        vec3 sunlight = transmittance_color(sun_position, moon_position);
                
        vec3 skylight = texture2(colortex8, project_sphere(vec3(0.0, 1.0, 0.0)) / ATMOSPHERE_LOD).xyz * pi;
             //skylight = mix(skylight, vec3(0.6) * luminance(skylight), wetness);

        vec3 scattering = direct * sunlight;
             scattering += indirect * skylight;

        return vec4(scattering, transmittance);
    }
#endif